(*
This file is part of the Kind verifier
* Copyright (c) 2007-2010 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


open Types
open Flags
open Exceptions
open Channels
open Globals
open Mpi
open Util


exception INV_BASE
let solver = Globals.my_solver
let toss x = () (* toss outputs *)



(********************************************************************)


let print_invariant b_eqs b_imps i_eqs i_imps depth=
  let buf = Buffer.create 100 in
     Buffer.add_string buf "(define ASSERTIONS-INV :: (-> _nat bool) (lambda ( _M::nat) (ite (= _M _base) true " ;
    Buffer.add_string buf "( and";
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.equ_invariant x);) b_eqs;
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.imp_invariant x);) b_imps;
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.int_eq_invariant x);) i_eqs;
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.int_imp_invariant x);) i_imps; 
     Buffer.add_string buf "))))";
     (Buffer.contents buf);;  

(********************************************************************)



let pr_func b_eqs b_imps i_eqs i_imps depth =
  pr "-- "; 
  print_int ((List.length b_imps) 
	      + 2*(List.length b_eqs) 
	      + 2*(List.length i_eqs) 
	      + (List.length i_imps) ) ;
  pr " implications. \n";
  pr "assert";
  let pos_l = Util.n_to_m 1 depth in
    List.iter (fun x-> pr " (true -> (") pos_l ; 
    List.map(fun x -> pr (Imp_graph.eq_org x); pr " \nand ";) b_eqs;
    List.map(fun x -> pr (Imp_graph.imp_org x); pr " \nand ";) b_imps;
    List.map(fun x -> pr (Imp_graph.int_eq_org x); pr " \nand ";) i_eqs;
    List.map(fun x -> pr (Imp_graph.int_imp_org x); pr " \nand ";) i_imps;
    pr " true ";
    List.iter (fun x-> pr "))") pos_l ; 
    pr ";\n" 


 
  let main fn =
  let defdoc,maxdepth,def_hash,pvars = Defgen.start_invariant_generator fn in
   
  Kind_support.print_stat_headers();
  
  Flags.invariant_bool := true;
  Flags.invariant_int := true;
  
   let start_time = ref 0.0 in
   let stop_time = ref 0.0 in
   let invariant_number = ref 0 in
     start_time := (wtime());
     
  (* Generate defs for new vars of equalities.*)
  let all_terms = Lus_convert_yc.get_all_def_terms () in
  let bool_sub_exprs, int_sub_exprs, float_sub_exprs = Sub_exprs.sub_exprs all_terms Expr_util.filter_sub_exprs in

  (* Generate new int and bool terms *)
  let new_bool_var_doc = New_vars.bool_newvar_defs bool_sub_exprs in
  let new_int_var_doc =  New_vars.int_newvar_defs int_sub_exprs in

  (* Creating bool and int equivalence classes. *)
  let bool_imp_graph =  new Imp_graph.bool_imp_graph (
      if (!Flags.invariant_bool) 
      then (New_vars.get_bool_nvrs ())
      else []
    )
  in
  let ints_partial_order = new Partial_order.partial_order (
      if (!Flags.invariant_int) 
      then New_vars.get_int_nvrs ()
      else []
    ) 
  in 


  let inv_gen_stop = ref false in
  let cc = solver#get#cc in
  let nvar = solver#get#k_position_string in
  let add = Kind_support.get_add() in
  let nstart = solver#get#step_base_string in
  let startdepth = maxdepth + 1 in
  let from_solver_ch, from_checker_ch, from_solver2_ch = Kind_support.setup_channels() in

  print_to_user("\n INCREMENTAL INV GEN PROC: Printing Lustre Program terms. \n"); 
  Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (defdoc^"\n");

  print_to_user("INCREMENTAL INV GEN PROC: Printing new terms. \n"); 
  print_to_both solver#get#push_string; (***** PUSH all new var defs*****)
  Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (new_bool_var_doc^"\n");
  Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (new_int_var_doc^"\n");

  print_to_user("INCREMENTAL INV GEN PROC: Printing Initialization. \n"); 
  print_to_both solver#get#push_string; (***** PUSH  Init of real terms*****)
  Kind_support.print_initialization def_hash startdepth add from_checker_ch;

  (* print new variables assertions. *)
  print_to_both solver#get#push_string; (***** PUSH Init of new vars*****)
  print_to_solver (New_vars.mk_nvr_eq_cmds (NUM 0));
 
  let k_provided = !Flags.k_incremental in
  let first_time_round = ref true in
  let error_inv = ref false in
  let stop_inv = ref false in
  let cur_depth = ref maxdepth  in
  let refinement_base = ref 0 in 
  
    Kind_support.set_last_level_asserted (startdepth-1);

 while not (!cur_depth = k_provided) && not (!Flags.stop_invariant) do 
      try
    
    let base_num = !cur_depth * add in
    let cur_depth_s = string_of_int !cur_depth in
    let refinement_base_s = string_of_int !refinement_base in 
      
      debug_step("INCREMENTAL INV GEN PROC: ASSERTING BASE CASE AT K = "^cur_depth_s^" -- Refinement round = "^refinement_base_s^"\n");
      
      print_to_solver solver#get#push_string; (***** PUSH  Candidate assertions *****)
      let bool_imp_doc = 
	if (!Flags.invariant_bool)
	then  bool_imp_graph#doc_of_asserts()
	else []
      in

      let int_imp_doc =  
	if (!Flags.invariant_int)
	then ints_partial_order#doc_of_asserts()
	else []
      in

      let equiv_imp_doc = Expr_util.mk_not_ands (bool_imp_doc@int_imp_doc) in 
	print_to_solver equiv_imp_doc;
	
	let assert_base = (F_EQ(POSITION_VAR(nstart),NUM(base_num),L_INT)) in
	  print_to_solver(Lus_convert_yc.simple_command_string (ASSERT assert_base)); 
	    print_to_solver(Lus_convert_yc.simple_command_string (QUERY(F_TRUE)));
	    
	    print_to_solver solver#get#done_string;
	    
	    print_to_user("INCREMENTAL INV GEN PROC: BASE - Checking entailement.\n");
	    
	    let out_base = solver#get#get_output from_solver_ch in
	      if (solver#get#result_is_unsat out_base) then
		begin (* base valid *)
		  print_to_user("INCREMENTAL INV GEN PROC: BASE CASE VALID AT K = "^cur_depth_s^"\n ");
		  print_to_solver solver#get#pop_string; (* PUSH 2 *)
		  let next_candidate_bool = bool_imp_graph#doc_of_asserts () in
		  let next_candidate_int = ints_partial_order#doc_of_asserts () in
		  raise INV_BASE;
		  
 end (* base valid *)
	  else if (solver#get#result_is_sat out_base) then
	    begin (* base invalid *)
     	      print_to_user("INCREMENTAL INV GEN PROC: BASE CASE INVALID AT K = "^cur_depth_s^"\n ");      

		let simulation_value_hash_base = 
		  solver#get#get_simulation_value_hash out_base print_to_solver from_solver_ch  in
		  
		let changed_base  = 
		  if (!Flags.invariant_int && !Flags.invariant_bool)
		  then (
		    let ra_base = bool_imp_graph#set_new_graph simulation_value_hash_base in
		    let rb_base = ints_partial_order#sort simulation_value_hash_base in
		      if (ra_base || rb_base ) then true else false )
		  else if (!Flags.invariant_bool)
		  then (bool_imp_graph#set_new_graph simulation_value_hash_base)
		  else if (!Flags.invariant_int)
		  then (ints_partial_order#sort simulation_value_hash_base)
		  else failwith "Error"
		in
		  print_to_solver solver#get#pop_string;
		  print_to_solver solver#get#safe_pop; (* PUSH 2 *) 
		  begin
		    stop_inv := not changed_base ; 
		    if(!stop_inv ) 
		    then (
		      pr (Expr_util.mk_not_ands (bool_imp_graph#doc_of_asserts() ));
		      pr "No changes while sat.\n";
		      failwith "global stop"; )
		  end;
		    refinement_base := !refinement_base + 1;		    
	    end (* base invalid *)
	  else
	    begin
              (* BASE check = ERROR *)
              if (solver#get#result_is_error out_base) then
		print_to_user_final ((Str.matched_string out_base)^"\n");
              print_to_user (cc^"SOLVER OUTPUT: "^out_base);
              error_inv := true;
              stop_inv := true
	    end;
 
with INV_BASE ->
  begin
    let step_inv_stop = ref false in
    let refinement_step = ref 0 in
      print_to_solver solver#get#pop_string; (* POP Init of new vars from base *)
      print_to_solver solver#get#pop_string; (* POP Init of real vars from base*)
      print_to_both solver#get#push_string;  (* PUSH candidate for the step case *)
      let positions = 
	List.map 
	  (fun x -> MINUS(POSITION_VAR(nvar),NUM(x))) 
	  (Util.n_to_m 1 (!cur_depth)) in		 
	List.iter (fun x -> 
		     Kind_support.def_assert_both1 def_hash "DEF" 
		       [PLUS(POSITION_VAR(nvar),NUM(x*add))]
		       !cur_depth from_checker_ch;) (Util.n_to_m 0 (!cur_depth));
	
	print_to_solver (New_vars.mk_nvr_eq_cmds (POSITION_VAR(nvar)));
	
	
	while not (!step_inv_stop) && not (!Flags.stop_invariant) do
	  let cur_depth_s = string_of_int !cur_depth in
	  let refinement_step_s = string_of_int !refinement_step in 
	    
print_to_user("INCREMENTAL INV GEN PROC: ASSERTING STEP CASE AT K = "^cur_depth_s^" -- Refinement round = "^refinement_step_s^"\n");
			
			   
	    print_to_both solver#get#push_string;  (* PUSH candidate for the step case *)  
	     let base_num = !cur_depth * add in
	       
	   if (!Flags.invariant_bool)
	   then 
	     List.iter 
	       (fun x-> print_to_solver x; )
	       (bool_imp_graph#doc_of_stepn_asserts positions); 
	  
	   if (!Flags.invariant_int)
	   then 
	      List.iter 
		(fun x-> print_to_solver x; )
		(ints_partial_order#doc_of_stepn_asserts positions); 
	      
	      let bool_imp_doc = 
		if !Flags.invariant_bool
		then (bool_imp_graph#doc_of_asserts ())
		else [] 
	      in

	      let int_imp_doc = 
		if !Flags.invariant_int
		then (ints_partial_order#doc_of_asserts ())
		else []
	      in
    
	      let imp_equiv_doc = Expr_util.mk_not_ands (bool_imp_doc@int_imp_doc) in   
		print_to_solver imp_equiv_doc;
		
		print_to_solver(Lus_convert_yc.simple_command_string (QUERY(F_TRUE)));
		print_to_solver solver#get#done_string;
		
		
		let out_step = solver#get#get_output from_solver_ch in
		  if (solver#get#result_is_unsat out_step) 
		  then(
		    print_to_user("INCREMENTAL INV GEN PROC: STEP CASE VALID K = "^cur_depth_s^"\n");
		    print_to_solver solver#get#pop_string;  (* POP Init of new vars for the step case + candidate*)
		    let invariant_number_s = string_of_int !invariant_number in
		    let b_eqs,b_imps =
		      if (!Flags.invariant_bool)
		      then bool_imp_graph#eq_imp_graph()
		      else [],[]
		    in
		    let int_eqs,int_imps =
		      if (!Flags.invariant_int)
		      then ints_partial_order#eq_imp()
		      else [],[]
		    in
		      Globals.inv_gen_time_stop := (wtime());
		
		      print_to_user("INCREMENTAL INV GEN PROC: Sending Invariant to STEP PROC. \n");
	
	pr_func b_eqs b_imps int_eqs int_imps maxdepth;
	      
	
              
		step_inv_stop := true;
		      
		  ) else if (solver#get#result_is_sat out_step) 
		  then(
		    print_to_user("INCREMENTAL INV GEN PROC: STEP CASE INVALID K = "^cur_depth_s^"\n");
		    
		    let simulation_value_hash = solver#get#get_simulation_value_hash out_step print_to_solver from_solver_ch  in
		    let changed = 
		      if (!Flags.invariant_bool && !Flags.invariant_int)
		      then (
			let ra = bool_imp_graph#set_new_graph simulation_value_hash in
			let rb = ints_partial_order#sort simulation_value_hash in
			  if (ra || rb) then true else false) 
		      else if (!Flags.invariant_bool)
		      then (bool_imp_graph#set_new_graph simulation_value_hash)
		      else if (!Flags.invariant_int)
		      then (ints_partial_order#sort simulation_value_hash )
		      else failwith "Error"
		    in
		
		  print_to_solver solver#get#pop_string; (* POP Init of new vars for the step case + candidate*)
		      
		      
		      refinement_step := !refinement_step + 1;
		  ) else ( (* error *)
		    if (solver#get#result_is_error out_step) then
		      print_to_user_final ((Str.matched_string out_step)^"\n");
		    print_to_user (cc^"SOLVER OUTPUT: "^out_step);
		    error_inv := true;
		    
		  );
		  
	done;
    end;
   first_time_round := false;
  print_to_solver solver#get#pop_string; (* POP Init of new vars for the step case + candidate*)
  print_to_both solver#get#push_string; (***** PUSH  Init of real terms*****)
  Kind_support.persistent_step_asserts_concrete 
    def_hash startdepth add (!cur_depth+1) from_checker_ch;
  print_to_both solver#get#push_string; (***** PUSH Init of new vars*****)
  print_to_solver (New_vars.mk_nvr_eq_cmds (NUM !cur_depth));
  refinement_base := 0;
  cur_depth := !cur_depth + 1 ;  
done
    



