(*
This file is part of the Kind verifier

* Copyright (c) 2007-2009 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(* k induction engine for lustre analysis *)
(* yices version *)

open Types
open Flags
open Exceptions
open Channels
open Globals

let solver = Globals.my_solver

let toss x = () (* toss outputs *)

(********************************************************************)
(* varhash: varname -> in/out (only) *)
(* allvars: varname -> in/out/local *)
let mainloop filename =
  let defdoc,maxdepth,def_hash,no_stateful,pvars = 
    Defgen.start filename 
  in
  (* if no stateful vars, then cancels loopfree mode *)
  if no_stateful then
    begin
      loopfree_mode := false;
      termination_check := false
    end;
  if (!termination_check || !loopfree_mode) && not !abs_mode then
    begin
      static_path_compression.(base_abstr) <- true;
      static_path_compression.(step_abstr) <- true
    end;

  let switch_to_static_compression = ref false in

(* note: only update abstr fields in def_hash *)

  if !best_first_path_mode then
    Abs.setup_best_scores def_hash pvars;

  if !print_dot_one then
    begin
      Deftable.print_defhash_graph def_hash;
      exit 0
    end;

  Kind_support.print_stat_headers();


  let cc = solver#get#cc in
  let nvar = solver#get#k_position_string in
  let add = Kind_support.get_add() in
  let nstart = solver#get#step_base_string in
  let startdepth = maxdepth + 1 in
  let from_solver_ch, from_checker_ch, from_solver2_ch = Kind_support.setup_channels() in



  print_to_user (cc^"Printing Definitions: \n");
  (* print definitions *)
  Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (defdoc^"\n");

  if !checker_mode then  print_to_checker solver#get#checker_setup_string;

  print_to_user (cc^"Printing Initial States:\n");
  (* print initial case *)

  print_to_both solver#get#push_string; (***** PUSH *****)
  Kind_support.print_initialization def_hash startdepth add from_checker_ch;


  let k_inc = ref startdepth in
  let maxsteps = max !maxloops (startdepth+1) in
  let firststep = ref true in

  Kind_refine.set_num_abstract ();
  Kind_refine.set_not_fully_refined();

  let refinement_pass = ref false in (* set to true if doing another pass 
                                        after refining *)

  let newly_refined_vars = ref [] in (* all vars refined this base k *)
  let collected_newly_refined_vars = ref [] in

  (* force refinements after increasing periods *)
  (* for sal_ind mode, we force refinement if k > the set size *)
  let force_step_refinement k =
    let dif = k - (Kind_refine.get_last_k_refined()) in
    (dif > !force_refinement) || (k >= !force_refinement && k = maxsteps-1)
  in

   (* all vars refined in base since last induction step *)
  let step_count = ref startdepth in
  (* make sure the count doesn't go off *)
  let do_this_k_step k =
    incr step_count;
    if !step_count >= !k_step_size then
      begin
        step_count := 0;
        true
      end
    else
      false
  in
  (* should be true if we will do a k step next time *)
  let do_term_check_now k =
    !step_count >= !k_step_size - 1
  in

  Kind_support.set_last_level_asserted (startdepth-1);


  (* from 1 to k-1 *)
  let first_time_n_values = 
    (Array.to_list (Array.init (!k_inc-1) (fun x -> x+1))) 
  in

  let final_position = 0 in
  (* this does not include 0 : 1 to k-1 *)
(*  while !k_inc < maxsteps && not !Globals.stop do Yeting for testing only*)
    while  not !Globals.stop do
    let k = !k_inc in
    let k1 = k-1 in
    let k_s = string_of_int k in

    (* Checking BASE *)
    print_to_user ("\n"^cc^"Checking k="^k_s^" BASE CASE\n");

    if not !refinement_pass then
      begin (* not refinement_pass *)

debug_to_user "not refinement_pass";

        (* handle assertions *)
	
        if Lus_assertions.assertions_present() then
          begin (* handle assertions *)
	    print_to_user  (cc^"Asserting assertions.\n");
            print_to_solver (
                      Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED(solver#get#assertions,
(*                            [PLUS(POSITION_VAR(nvar),NUM(add*k1))])*)
				 [NUM(add*k)])
                           )));
	     print_to_user  (cc^"Normal assertions.\n");
            print_to_solver2 (
                      Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED(solver#get#assertions,
                            [PLUS(POSITION_VAR(nvar),NUM(add*k))])
                           )));
	    print_to_user  (cc^"Extra assertions.\n");
            print_to_solver2 (
                      Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED(solver#get#assertions,
                            [PLUS(POSITION_VAR(nvar),NUM(k+1))])
                           )));
            if !Globals.firsttime then
              for i= 0 to k do
                print_to_solver (
                      Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED(solver#get#assertions,
                             [NUM(add*i)]))));
		 print_to_user  (cc^"Normal assertions.\n");
                print_to_solver2 (
                      Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED(solver#get#assertions,
                            [PLUS(POSITION_VAR(nvar),NUM(add*i))])
                           )));
		print_to_user  (cc^"Extra assertions.\n");
		print_to_solver2 (
                      Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED(solver#get#assertions,
                            [PLUS(POSITION_VAR(nvar),NUM(i+1))])
                           )));
              done;
		print_to_solver  (cc^"Asserting assertions --  DONE\n");
          end; (* handle assertions *)

        print_to_solver solver#get#push_string; (***** PUSH *****)


        (* properties are defined & asserted normally *)
        (* assert the properties here*)
        let impl_premise = 
          F_EQ(POSITION_VAR(nstart),NUM(add*k1),L_INT)
        in
        let impl_result = 
          if !Globals.firsttime && k > 1 then
            begin
              List.fold_left (fun acc x -> 
                F_AND(F_PRED("P",[NUM(add*x)]), acc)) (F_PRED ("P",[NUM(0)])) 
                first_time_n_values
            end
          else
            F_PRED("P",[NUM(final_position)])
        in

        print_to_solver (
          Lus_convert_yc.simple_command_string (
            ASSERT(
               F_NOT(F_IMPL(impl_premise,impl_result))
            )
          )
        );
        print_to_solver solver#get#safe_push; (* PUSH 2a *)
        print_to_solver (
          Lus_convert_yc.simple_command_string (
            QUERY(F_TRUE)))
      end (* not refinement_pass *)
    else
      begin
        print_to_solver solver#get#safe_push; (* PUSH 2b *)
        print_to_solver (Lus_convert_yc.simple_command_string 
          (QUERY(F_TRUE)))
      end;




    (* BASE check *)
    print_to_solver solver#get#done_string;
    (* finished BASE check *)
    
    let out1 = solver#get#get_output from_solver_ch in


    if (solver#get#result_is_unsat out1) then
      begin (* base valid *)
        (* BASE check = VALID *)
        print_to_solver solver#get#safe_pop; (* PUSH 2 *)
        debug_to_user "base valid";
        print_to_solver solver#get#pop_string; (***** POP *****)
        (* re-assert the base defs for all newly refined vars *)
        List.iter (fun y ->
          Kind_support.print_initialization_single def_hash startdepth add y;
          for i=startdepth to k-1 do
            Kind_support.persistent_step_single_assert def_hash startdepth 
              add BASE i y
          done
        ) !newly_refined_vars;

        (* add any newly refined vars to those from last step *)
        if !do_k_induction then
          collected_newly_refined_vars := 
            List.rev_append (!collected_newly_refined_vars) 
              !newly_refined_vars;

        newly_refined_vars := [];

          print_to_user (cc^k_s^" BASE CASE is Valid. Continuing search.\n");

        if !Globals.firsttime then
          begin
            List.iter (fun x -> print_to_solver
                      (Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED("P",[NUM(add*(x+1))])
                           )))
            ) (0::first_time_n_values)
          end
        else
          print_to_solver 
                      (Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED("P",[NUM(add*k)])
                           )));


        if !do_k_induction && do_this_k_step k then
          begin (* do step *)
          (* permantly add newly refined vars to step solver *)
print_to_user (cc^"Beginning STEP CASE\n");
            (* we only go to k- step size *)
            List.iter (fun y ->
              Kind_support.def_assert_initial def_hash "DEF" add y;
              for i=startdepth to k - !k_step_size do
                 Kind_support.persistent_step_single_assert def_hash 
                   startdepth add STEP i y
              done
            ) !collected_newly_refined_vars;

            collected_newly_refined_vars := [];

            let do_range = !Globals.firsttime || !k_step_size > 1 in

            if !loopfree_mode && static_path_compression.(step_abstr) then
              begin
                let startpoint =
                  if !firststep then startdepth
                  else k - startdepth - !k_step_size
                in
                Kind_compression.general_compression_constraint startdepth 
                  add print_to_solver2 startpoint (k1-startdepth) nvar 
              end;

            let perform_step_check = ref true in

            let previous_static_compression = 
              ref static_path_compression.(step_abstr) 
            in
            let initialize_step = ref true in
            let step_refined_vars = ref [] in

            while !perform_step_check do 
            (* step loop *)
              perform_step_check := false;
            

              (* Checkng STEP *)

              if !initialize_step || not !previous_static_compression then
                begin (* setup step loop *)
                  initialize_step := false;
                  debug_to_user "setup step loop: initilize step";

                  if !loopfree_mode && 
                    not static_path_compression.(step_abstr) then
                    begin
                      try
                        if !switch_to_static_compression then
                          begin
                            switch_to_static_compression := false;
                            raise AllStatefulVarsRefined
                          end;

                        let header_string = 
                          solver#get#push_string^
                            (Lus_convert_yc.yc_state_vars_string_refined 
                            step_abstr)^"\n"
                        in
                        print_to_checker solver#get#push_string;
debug_to_user ";4c";
                        print_to_solver2 header_string;
                        Kind_compression.abstracted_compression_constraint 
                          startdepth add print_to_solver2 0 (k1-startdepth) 
                          nvar ;
                      with NoStatefulVars -> 
debug_to_user ";4e";
                       print_to_both2 solver#get#push_string(***** PUSH *****) 
                      | AllStatefulVarsRefined ->
                        static_path_compression.(step_abstr) <- true;
                        Kind_compression.general_compression_constraint 
                          startdepth add print_to_solver2 0 (k1-startdepth) 
                          nvar ;
                        previous_static_compression := true;
debug_to_user ";4f";
                        print_to_both2 solver#get#push_string(***** PUSH *****)
                    end
                  else
                    begin
                      print_to_both2 solver#get#push_string;(***** PUSH *****)
debug_to_user ";4b";
                      previous_static_compression := true
                    end;

                  (* generic terms *)
                  (* temporarily add them since the last startdepth *)
                  (* takes care of ite elim *)
                  for i=k-startdepth to k-1 do
                    if !var_defs_mode then
		      begin
		      	print_to_user(cc^"Normal steps -- Vars defs mode.\n");
			Kind_support.def_assert_both2 def_hash "DEF" 
                          [PLUS(POSITION_VAR(nvar),NUM(i*add))]
                          k from_checker_ch;
			 print_to_user(cc^"Extra backwards reachability -- Vars defs mode.\n");
			 Kind_support.def_assert_both2 def_hash "DEF" 
                          [PLUS(POSITION_VAR(nvar),NUM(i+1))]
                          k from_checker_ch;
			    end
                    else
                      begin
			print_to_user(cc^"Normal steps -- No Vars defs mode.\n");
(**)                    print_to_both2 
                         (Lus_convert_yc.simple_command_string
                         (ASSERT(
                         F_PRED("DEF",[PLUS(POSITION_VAR(nvar),NUM(add*i))])
                         )));
			  print_to_user(cc^"Extra backwards reachability -- No Vars defs mode.\n");
			   print_to_both2 
                         (Lus_convert_yc.simple_command_string
                         (ASSERT(
                         F_PRED("DEF",[PLUS(POSITION_VAR(nvar),NUM(i+1))])
                         )))
                      end
                  done;
                  (* temporarily print all defs up from the last step up to 
                   k-startdepth-1
                   this includes recently refined ones
                   remember that on the previous step we kept one step size *)
                  (* all skipped defs up to k-startdepth-1 *)
                  if !k_step_size > 1 then
                    begin (* setup step loop *)
                      (* fill in for k_step_size > 1 *)
                      let startpoint,endpoint = 
                        if !firststep then
                          0,k-startdepth-1
                        else
                          k - !k_step_size - startdepth + 1, k-startdepth-1
                      in 
                      let defname = "DEF" in
                      for i=startpoint to endpoint do
                        if !var_defs_mode then
			  begin
(**)                      Kind_support.def_assert_both2 def_hash defname 
                            [PLUS(POSITION_VAR(nvar),NUM(i*add))]
                            k from_checker_ch
			      end
                        else
                          begin
(**)                        print_to_both2 
                              (Lus_convert_yc.simple_command_string
                              (ASSERT(
                               F_PRED(defname,[PLUS(POSITION_VAR(nvar),
                                 NUM(add*i))])
                              )))
                          end
                      done;
                    end;

                  if do_range then
                    begin
                      let startpoint = 
                        if !firststep then 1
                        else k - !k_step_size + 1
                      in
                      for i=startpoint to k-1 do
                        print_to_solver2 
                          (Lus_convert_yc.simple_command_string
                          (ASSERT(
                           F_PRED("P",[PLUS(POSITION_VAR(nvar),NUM(add*i))])
                           )))
                      done
                    end;

                  print_to_solver2
                    (Lus_convert_yc.simple_command_string
                    (ASSERT(
                      F_NOT(F_IMPL(
                        F_EQ(POSITION_VAR(nstart),NUM(add*k1),L_INT),
                        F_PRED("P",[POSITION_VAR(nvar)])
                      ))
                    )));


                end (* setup step loop *)
              else
                begin (* alt setup step loop *)
debug_to_user "setup step loop: no initilize";

                  print_to_checker solver#get#push_string;

                  for i=k-startdepth to k-1 do
                    if !var_defs_mode then
		      begin
			print_to_user(cc^"Normal steps -- Vars defs mode.\n");
			Kind_support.def_assert_both2 def_hash "DEF" 
                              [PLUS(POSITION_VAR(nvar),NUM(i*add))]
                              k from_checker_ch;
			print_to_user(cc^"Extra backwards reachability -- Vars defs mode.\n");
			Kind_support.def_assert_both2 def_hash "DEF" 
                          [PLUS(POSITION_VAR(nvar),NUM(i+1))]
                          k from_checker_ch;
			  end
                    else
                      begin
			print_to_user(cc^"Normal steps.\n");
		print_to_checker 
                      (Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED("DEF",[PLUS(POSITION_VAR(nvar),NUM(add*i))])
                           )));
		print_to_user(cc^"Extra backwards reachability.\n");
		print_to_checker 
                      (Lus_convert_yc.simple_command_string
                        (ASSERT(
                           F_PRED("DEF",[PLUS(POSITION_VAR(nvar),NUM(i+1))])
                           )))
                      end
                  done;
                  if !k_step_size > 1 then
                    begin
                      (* fill in for k_step_size > 1 *)
                      let startpoint,endpoint = 
                        if !firststep then
                          0,k-startdepth-1
                        else
                          k - !k_step_size - startdepth + 1, k-startdepth-1
                      in 
                      let defname = "DEF" in
                      for i=startpoint to endpoint do
                        if !var_defs_mode then
			  begin
			  print_to_user(cc^"Normal steps  -- Var defs mode + Step_size > 1 .\n");
(**)                        Kind_support.def_assert_both2 def_hash defname 
                              [PLUS(POSITION_VAR(nvar),NUM(i*add))]
                              k from_checker_ch;
			      print_to_user(cc^"Extra backwards reachability  -- Var defs mode + Step_size > 1 .\n");
(**)                        Kind_support.def_assert_both2 def_hash defname 
                              [PLUS(POSITION_VAR(nvar),NUM(i+1))]
				k from_checker_ch;
				end
                        else
                          begin
			    print_to_user(cc^"Normal steps  -- Step_size > 1 .\n");
(**)                        print_to_checker 
                              (Lus_convert_yc.simple_command_string
                              (ASSERT(
                               F_PRED(defname,[PLUS(POSITION_VAR(nvar),
                               NUM(add*i))])
                               )));
				print_to_user(cc^"Extra backwards reachability  -- Step_size > 1 .\n");
(**)                        print_to_checker 
                              (Lus_convert_yc.simple_command_string
                              (ASSERT(
                               F_PRED(defname,[PLUS(POSITION_VAR(nvar),
                               NUM(i+1))])
                               )))
                          end
                      done
                    end
                end; (* alt setup step loop *)

                (* STEP check *)
                print_to_user (cc^"Checking k="^k_s^" step\n");

                print_to_solver2 solver#get#safe_push; (* SAFE PUSH 1 *)

                print_to_solver2 
                      (Lus_convert_yc.simple_command_string
                        (QUERY(F_TRUE)));

                print_to_solver2 solver#get#done_string;


                (* finished STEP check *)
                let out2 = solver#get#get_output from_solver2_ch in
                if (solver#get#result_is_unsat out2) then
                  begin (* step valid *)
                    (* STEP check = VALID *)

                    print_to_solver2 solver#get#safe_pop; (* SAFE PUSH 1 *)
                    if !check_compression && !loopfree_mode 
                      && not static_path_compression.(step_abstr) then
                      begin
                        print_to_user 
                          (cc^"Re-checking with full compression.\n");
                        switch_to_static_compression := true;
                        print_to_both2 solver#get#pop_string;
                        (* add the static def here *)
                        initialize_step := true;
                        perform_step_check := true
                      end
                    else
                      begin
                        print_to_user_final (cc^k_s^" step is Valid.\n");
                        print_to_user_final 
                          (cc^"+++ The property holds. +++\n");
                        Globals.stop := true
                      end
                  end (* step valid *)
                else if (solver#get#result_is_sat out2 
                  || solver#get#result_is_unknown out2) then
                  begin (* step invalid *)
                    (* STEP check = INVALID or UNKNOWN *)
                    let go_to_next =
                      (* if we refined, check this level again *)
                      if !abs_mode &&
                       Kind_refine.is_not_fully_refined step_abstr && 
                       (Kind_refine.check_step_refinement k) &&
                       (solver#get#result_is_sat out2) then
                       begin (* step invalid abs *)
                         let foundvars = 
                           solver#get#get_countermodel out2 print_to_solver2 
                             from_solver2_ch 
                         in
                         let 
                           refinement_result = 
                           (* may POP stack *)
debug_to_user "check1";
                           print_to_solver2 solver#get#safe_pop;(* SAFE PUSH 1 *)
                           Kind_refine.check_and_refine_abstraction 
                             from_checker_ch def_hash startdepth add STEP 
                             foundvars pvars k
                         in
                         let previous_step_k = k-1 in
                         match refinement_result with
                           CHECK_IS_VALID ->
                             if force_step_refinement k then
                               begin
                                 let ids_steps = 
                                   List.map (fun x -> (x,0)) pvars 
                                 in
                                 print_to_user (cc^"Forcing refinement.\n");
                                 (* may POP stack *)
                                 let xs = 
                                   Kind_refine.refine_abstraction 
                                     from_checker_ch def_hash startdepth add 
                                     false STEP foundvars k ids_steps
                                 in
                                 step_refined_vars := List.rev_append 
                                   !step_refined_vars xs;

                                 if !previous_static_compression then
                                   print_to_checker solver#get#pop_string
                                 else
                                   begin
                                     initialize_step := true;
                                     print_to_both2 solver#get#pop_string;
                                   end;
debug_to_user("CHECK_IS_VALID");
                                 (* assert new vars *)
                                 List.iter (fun y ->
                                   (* assert up to startdepth *)
                                   Kind_support.def_assert_initial def_hash 
                                     "DEF" add y;
                                   Kind_support.print_initialization_single 
                                     def_hash startdepth add y;
                                (* assert from startdepth to previous_step_k *)
                                   for i=startdepth to previous_step_k do
                                     Kind_support.persistent_step_single_assert
                                       def_hash startdepth add STEP i y
                                   done;
                                   (* assert in base up to k-1 *)
                                   if !Flags.only_1_abstraction then
                                   for i=startdepth to k-1 do
                                     Kind_support.persistent_step_single_assert
                                       def_hash startdepth add BASE i y
                                   done
                                 ) xs;
                                 false
                               end
                             else
                               true
                           | CHECK_IS_INVALID x ->
                             step_refined_vars := List.rev_append 
                               !step_refined_vars x;

                             if !previous_static_compression then
                               print_to_checker solver#get#pop_string
                             else
                               begin
                                 initialize_step := true;
                                 print_to_both2 solver#get#pop_string
                               end;

debug_to_user("CHECK_IS_INVALID");
                             (* assert new vars up to previous step *)
                             List.iter (fun y ->
                               (* assert up to startdepth *)
                               Kind_support.def_assert_initial def_hash "DEF" 
                                 add y;
                               Kind_support.print_initialization_single 
                                 def_hash startdepth add y;
                               for i=startdepth to previous_step_k do
                               (* assert from startdepth to previous_step_k *)
                                 Kind_support.persistent_step_single_assert 
                                   def_hash startdepth add STEP i y
                               done;
                               (* assert in base up to k-1 *)
                               if !Flags.only_1_abstraction then
                               for i=startdepth to k-1 do
                                 Kind_support.persistent_step_single_assert 
                                   def_hash startdepth add BASE i y
                               done
                             ) x;
                             false (* also POP stack *)
                       end (* step invalid abs *)
                     else
                       begin
                         print_to_solver2 solver#get#safe_pop;(* SAFE PUSH 1 *)
                         true
                       end
                    in (* go_to_next done *)
                    if go_to_next then
                      begin (* step invalid continue *)
                        if solver#get#result_is_unknown out2 then
                          print_to_user (cc^"Unknown in "^k_s
                            ^" step. Continuing search.\n")
                        else
                          print_to_user (cc^k_s
                            ^" step is Invalid. Continuing search.\n");

(**)                    print_to_both2 solver#get#pop_string; (***** POP *****)

                        if do_range then
                          begin
                            let startpoint = 
                              if !firststep then 1
                              else k - !k_step_size + 1
                            in
                            for i=startpoint to k-1 do
                              print_to_solver2 
                               (Lus_convert_yc.simple_command_string
                               (ASSERT(
                               F_PRED("P",[PLUS(POSITION_VAR(nvar),NUM(add*i))])
                               )))
                            done
                          end;

                          print_to_solver2 
                            (Lus_convert_yc.simple_command_string
                            (ASSERT(
                             F_PRED("P",[PLUS(POSITION_VAR(nvar),NUM(add*k))])
                             )));

                          (* do permanent step asserts *)
                          let startpoint =
                            if !firststep then startdepth
                            else k - !k_step_size + 1
                          in
                          (* assert all from stepsize to k-1 *)
                          for i=startpoint to k - 1 do
print_to_both2 (";step asserts symbolic "^(string_of_int i)^":\n");
                            Kind_support.persistent_step_asserts_symbolic 
                              def_hash startdepth add i from_checker_ch
                          done;
                          (* presistent asserts of refined vars *)
                          List.iter (fun y ->
                            Kind_support.def_assert_initial def_hash "DEF" add y;
                            Kind_support.print_initialization_single def_hash 
                              startdepth add y;
                            if !Flags.only_1_abstraction then
                            for i=startdepth to k-1 do
                              Kind_support.persistent_step_single_assert 
                                def_hash startdepth add BASE i y
                            done;
                            for i=startdepth to startpoint-1 do
                              Kind_support.persistent_step_single_assert 
                                def_hash startdepth add STEP i y
                            done
                          ) !step_refined_vars;
                          (* assert all at k *)
                          Kind_support.persistent_step_asserts_concrete 
                            def_hash startdepth add k from_checker_ch;
                          Kind_support.persistent_step_asserts_symbolic 
                            def_hash startdepth add k from_checker_ch;
                          Globals.firsttime := false;
                          firststep := false;
                          incr k_inc
                      end (* step invalid continue *)
                    else (* re-do step *)
                      begin
                        perform_step_check := true
                      end
                  end (* step invalid *)
                else
                  begin
                    (* STEP check = ERROR *)
                    print_to_user (cc^"Abort in "^k_s^" step.\n");
                    if (solver#get#result_is_error out2) then
                      print_to_user_final ((Str.matched_string out2)^"\n");

                    print_to_user (cc^"OUTPUT: "^out2);
                    Globals.error := true;
                    Globals.stop := true
                  end
            done; (* step loop *)

            refinement_pass := false;
            (* term check may break this *)
            if !incr_k_step_size then
              incr k_step_size
          end (* do step *)
        else (* BMC only *)
          begin
            (* assert at k *)
(**)        Kind_support.persistent_step_asserts_concrete def_hash startdepth 
              add k from_checker_ch;
            refinement_pass := false;
            Globals.firsttime := false;
            incr k_inc
          end
      end (* base valid *)
    else if (solver#get#result_is_sat out1) then
      begin (* base invlaid *)
        (* BASE check = INVALID *)
debug_to_user "base invalid";
        let basestep = 
          if !Globals.firsttime then
            BASE_INIT
          else
            BASE
        in
        let foundvars = 
          solver#get#get_countermodel out1 print_to_solver from_solver_ch 
        in
debug_to_user ("checking absmode "^(string_of_int (Kind_refine.get_num_abstract base_abstr)));
        print_to_solver solver#get#safe_pop; (* PUSH 2 *)
        if !abs_mode && Kind_refine.is_not_fully_refined base_abstr then
          begin
            (* may POP stack *)
debug_to_user "check2";
            match Kind_refine.check_and_refine_abstraction from_checker_ch 
              def_hash startdepth add basestep foundvars pvars k with
                  CHECK_IS_VALID -> () (* continue *)
                | CHECK_IS_INVALID xs ->
                   (* assert new vars *)
                   List.iter (fun y ->
                     (* assert just y up to base step *)
                     Kind_support.print_initialization_single def_hash 
                       startdepth add y;
                     (* assert just y from startdepth to k-1 *)
                     for i=startdepth to k-1 do
                       Kind_support.persistent_step_single_assert def_hash 
                         startdepth add basestep i y
                     done
                   ) xs;
                   (* note these new vars *)
                   newly_refined_vars := 
                     List.rev_append (!newly_refined_vars) xs;
                   refinement_pass := true
(***)
          end
        else
          begin
            print_to_user_final (cc^k_s^" base is Invalid.\n");
            if !checker_mode then
              toss(Kind_support.query_checker from_checker_ch foundvars basestep k1);

            Kind_support.print_countermodel foundvars 0 k1;
            Globals.stop := true
          end
      end (* base invalid *)
    else
      begin
        (* BASE check = ERROR *)
        print_to_user_final (cc^"Abort in "^k_s^" base due to incompleteness or error.\n");
        if (solver#get#result_is_error out1) then
          print_to_user_final ((Str.matched_string out1)^"\n");
        print_to_user (cc^"SOLVER OUTPUT: "^out1);
        Globals.error := true;
        Globals.stop := true
      end;

    if not !Globals.stop && !termination_check && not !refinement_pass 
      && k >= startdepth 
      && ((do_term_check_now k) || not !interleave_termination_checks)
    then
      Kind_compression.termination_check def_hash startdepth !k_step_size 
        !firststep add newly_refined_vars k from_solver_ch from_checker_ch;

  done; (* main loop *)

  if not !Globals.stop then
    print_to_user_final (cc^"+++ Unknown (max iterations reached) +++\n");
  Kind_refine.print_stats def_hash;
  print_to_user_final (cc^"Done.\n")


