(*
This file is part of the Kind verifier

* Copyright (c) 2007-2012 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


(** Incremental and Parallel Kind for verification of multiple properties.  Main entry point. *)

(** 
@author Temesghen Kahsai

*)

open Types
open Exceptions
open Channels
open Globals


let solver = Globals.my_solver

(********************************************************************)
let print_copyright () =
  print_string (
   "\nIncremental and parallel K-Induction-based model checker (version "^Globals.version
  ^", build "^Globals.build^")\n"
  ^"* Copyright (c) 2007-2011 by the Board of Trustees of the University of Iowa.\n"
  ^"* All rights reserved.\n"
  ^"* \n"
  ^"* Redistribution and use in source and binary forms, with or without\n"
  ^"* modification, are permitted provided that the following conditions are met:\n"
  ^"*     * Redistributions of source code must retain the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer.\n"
  ^"*     * Redistributions in binary form must reproduce the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer in the\n"
  ^"*       documentation and/or other materials provided with the distribution.\n"
  ^"*     * Neither the name of the University of Iowa, nor the\n"
  ^"*       names of its contributors may be used to endorse or promote products\n"
  ^"*       derived from this software without specific prior written permission.\n"
  ^"*\n"
(*  
  ^"* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY\n"
  ^"* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n"
  ^"* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n"
  ^"* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY\n"
  ^"* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n"
  ^"* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n"
  ^"* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n"
  ^"* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"
  ^"* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"
  ^"* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"
*)
  )


(********************************************************************)
let catch_alarm x =
  print_to_user_final (solver#get#cc^"Signal caught: Timeout. Terminating.\n");
  exit x

let catch_term x =
  print_to_user_final (solver#get#cc^"Signal caught: Memory out. Terminating.\n");
  exit x

let catch_pipe x = 
  print_to_user_final (solver#get#cc^"Signal caught: Broken pipe. The solver has unexpectedly closed.  Terminating.\n"
  ^(match !Flags.set_my_solver with
        YICES -> ""
      | YICES_WRAPPER -> 
          solver#get#cc^"It is possible a formula exceeded the interactive yices buffer.\n"
      | CVC3 -> (if !Flags.abs_mode then 
                   solver#get#cc^"It is possible CVC3 failed to instantiate a counterexample.\n"
                   ^solver#get#cc^"Try re-running with abstraction turned off.\n" else "")
      | Z3 -> ""
      | CVC4 -> (if !Flags.abs_mode then 
                   solver#get#cc^"It is possible CVC4 failed to instantiate a counterexample.\n"
                   ^solver#get#cc^"Try re-running with abstraction turned off.\n" else ""))
		      );
  exit x
			 

module PS = Proc_scheduler.Dispatch (* The process scheduler *) 

(********************************************************************)
let inv_gen = ref false
let kind_ai = ref false

let start fn =
  begin
    match !Flags.set_my_solver with
        YICES ->
          solver#set (new Solver_yices_cmdline.solver_yices_cmdline)
      | YICES_WRAPPER -> 
	  begin
	    (* In case the yices warpper is not available switch to yices command line version *)
	    if(Solvers_path.yicesw_path = "no") then
	      (
		solver#set (new Solver_yices_cmdline.solver_yices_cmdline)
	      )else(
		solver#set (new Solver_yices.solver_yices)
	      );
	  end            
      | CVC3 -> 
          solver#set (new Solver_cvc3.solver_cvc3)

      | CVC4 -> 
	  solver#set (new Solver_cvc4.solver_cvc4)

      | Z3 ->
	  solver#set (new Solver_z3.solver_z3)
      end;

  (* Store the commmands sent to the solver in auxilary files, i.e. for debugging purposes.*)
      if !Flags.do_scratch then(
	  if proc_size = 2 then (
          base_ch := open_out (fn^"."^solver#get#file_extension^"_base");
          induct_ch := open_out (fn^"."^solver#get#file_extension^"_induct");
	  )else(
	    base_ch := open_out (fn^"."^solver#get#file_extension^"_base");
            induct_ch := open_out (fn^"."^solver#get#file_extension^"_induct");
	    inv_ch := open_out (fn^"."^solver#get#file_extension^"_inv");
	  );
      ); 

      (* If XML is activated generate the output in a file*)
      if !Flags.xml then (
	xml_ch := open_out (fn^".xml");
	print_xml (Kind_support.print_header_xml fn);
      );
      (
	match !inv_gen, !kind_ai with
	  | true, true -> (
	    inv_gen_proc := 2;
	    kind_ai_proc := 3
	  )
	  | false, true -> (
	    inv_gen_proc := -1;
	    kind_ai_proc := 2
	  )
	  | true, false -> (
	    inv_gen_proc := 2;
	    kind_ai_proc := -1
	  )
	  | false, false -> ()
      );
      (* Schedule processes *)
      PS.schedule fn proc_size;;

(********************************************************************)
let _ =
  let usage_msg = "PKIND -- Incremental and Parallel k-induction model checker for the verification of multiple properties Lustre programs. 
                   \nUsage: mpiexec -np [2 or 3] pkind  [options] input_file" in
  let speclist = [
		   ("-with-inv-gen",
                    Arg.Set inv_gen,
                    "\tActivate INVARIANT GENERATION based on a template. It will generate invariants of type BOOL and INT."
                   );

		   ("-with-kind-ai",
                    Arg.Set kind_ai,
                    "\tActive INVARIANT GENERATION based on abstract interpretation. KIND-AI  must be installed and configures with './configure'."
                   );

                   ("-node", Arg.String (fun x -> Flags.user_specified_main_node_name := x),
		     "\tCONTROL: User specified main node (default: the last node that contains a PROPERTY)");

                   ("-n",
                    Arg.Int (fun x -> Flags.maxloops := x),
                    "\tCONTROL: Stop after n iterations (default 200)"
                   );

		   ("-no_multi",
                    Arg.Unit (fun () -> Flags.no_multi := true),
                    "\tCONTROL: Disables the verification of multiple properties in an incremental fascion. (default: true)"
		   );

		   ("-timeout",
                    Arg.Float (fun x -> Flags.timeout := x),
                    "\tCONTROL: For setting a timeout (default: 100.00)"
                   );
         		   
		   ("-k_incremental",
                    Arg.Int (fun x -> Flags.k_incremental := x+1),
                    "\tINVARIANT GENERATION: Generate incremental invariants until n iterations (default: k=200)"
                   );
		   
		   ("-no_incremental",
                    Arg.Unit (fun () -> Flags.incremental := false),
                    "\tINVARIANT GENERATION: Generate non incremental invariants. (default: generate incremental invariants)"
                   );

		   ("-range", Arg.Int (fun x -> Flags.range := x ),
		    "\tINVARIANT GENERATION: Maximum range for which to generate candidates for mode variables (default: max range = 10)");
		   
		   ("-no_mode_inv",
                    Arg.Unit (fun () -> Flags.mode_inv := false),
                    "\tINVARIANT GENERATION: Don not generate invariants for mode variables. (default: Generate invariants for mode variables)"
                   );

                   ("-mpi_abort",
                    Arg.Unit (fun () -> Flags.mpi_abort := true),
                    "\tPARALLELISM: It will abort the mpi environment. Useful when the solver returns with an answer but the environment is not terminated yet. (default false)"
                   );
                   
		   ("-cvc3",
                    Arg.Unit (fun () -> Flags.set_my_solver := CVC3),
                    "\tSOLVER: Use CVC3 as solver (default: yicesw)"
                   );

		   ("-cvc4",
                    Arg.Unit (fun () -> Flags.set_my_solver := CVC4),
                    "\tSOLVER: Use CVC4 as solver [Work in progress] (default: yicesw)"
                   );
                   
		   ("-yices_exec",
                    Arg.Unit (fun () -> Flags.set_my_solver := YICES),
                    "\tSOLVER: Use Yices executable as solver (default: yices)"
                   );
  
		   ("-solver_path",
                    Arg.String (fun x -> Flags.solver_path := x;
				  Flags.gui:=true),
                    "\tSOLVER: Specify the path to the solver. Used for KIND GUI. (default: specified by ./configure)"
                   );
         
		   ("-compression",
                    Arg.Unit (fun () -> Flags.loopfree_mode := true),
                    "\tCOMPR: Include path compression but not term check (default: false)"
                   );
		   
		   ("-loop",
                    Arg.Unit (fun () -> Flags.loopfree_mode := true;
                                Flags.termination_check := true),
                    "\tCOMPR: Include loopfree constraint formulas + term check (default: false)"
                   );
		   
		   ("-flags",
                    Arg.String (fun x -> Flags.solverflags := x),
                    "\tSOLVER: Pass string to solver cammand line (default: n/a)"
                   );

		   ("-xml",
		    Arg.Unit (fun x -> Flags.xml := true),
                    "\tOUTPUT: Generate KIND result in XML format (default: false)"
                   );
		   
                   ("-quiet",
                    Arg.Unit (fun () -> Flags.loud := false;
                                        Flags.final_cex_loud := false),
                    "\tOUTPUT: Only return final answer (default: true)"
                   );
                   
		   ("-verbose",
                    Arg.Unit (fun () -> Flags.loud := true;
                                        Flags.final_cex_loud := true),
                    "\tOUTPUT: Return more details (default: false)"
                   );

		   ("-induct_cex",
                    Arg.Unit (fun () -> Flags.inductive_cs := true),
                    "\tOUTPUT: Prints the current inductive counterexample (default: false)"
                   );

		    ("-scratch",
                    Arg.Unit (fun () -> Flags.do_scratch := true),
                    "\tOUTPUT: Produce files for debugging purposes. This option will produce 3 files (*_base, *_induct, *_inv). (default: Don't log to disk)"
                   );

                   ("-only_inputs",
                    Arg.Unit (fun () -> Flags.print_all_vars := false),
                    "\tOUTPUT: Only print inputs in counterexamples (default: false)"
                   );

                   ("-dot",
                    Arg.Unit (fun () -> Flags.print_dot_one := true;
                                        Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.loud := false),
                    "\tOUTPUT: Prints dot graph of dependencies (default: false)"
                   );

                   ("-dotall",
                    Arg.Unit (fun () -> Flags.print_dot_all := true;
                                        Flags.abs_mode := true;
                                        Flags.var_defs_mode := true),
                    "\tOUTPUT: Outputs all abstraction dependency graphs to graphXXX.dot (default: false)"
                   );  
    
                   ("-version",
                      Arg.Unit (fun () -> print_copyright()),
                      "Print version & copyright information"
                   )
                 ]
  in
  Arg.parse speclist start usage_msg 


